// Copyright 2018 Consorsbank
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// tag::Using[]
using System;
using System.Collections.Generic;
using System.IO;
using Grpc.Core;
using Com.Consorsbank.Module.Tapi.Grpc;
using System.Threading.Tasks;
using System.Threading;
using Microsoft.CSharp.RuntimeBinder;

// end::Using[]

namespace TestClient
{
    // tag::Main[]
    class Program : IDisposable
    {
        // tag::Empty[]
        private static readonly Empty empty = new Empty();
        // end::Empty[]

        private readonly Channel channel;
        // tag::stockExchangeServiceClient[]
        private readonly StockExchangeService.StockExchangeServiceClient stockExchangeServiceClient;
        // end::stockExchangeServiceClient[]
        // tag::depotServiceClient[]
        private readonly DepotService.DepotServiceClient depotServiceClient;
        // end::depotServiceClient[]
        // tag::securityServiceClient[]
        private readonly SecurityService.SecurityServiceClient securityServiceClient;
        // end::securityServiceClient[]
        // tag::orderServiceClient[]
        private readonly OrderService.OrderServiceClient orderServiceClient;
        // end::orderServiceClient[]

        // tag::Login[]
        private readonly string _accessToken = "";

        // end::Login[]
        // end::Main[]

        // tag::StreamMarketDataAsync[]
        private async Task StreamMarketDataAsync()
        {

            MarketDataStreamer marketDataStreamer = 
                new MarketDataStreamer(securityServiceClient); // <1>

            SecurityMarketDataRequest marketDataRequestTRG = // <2>
                marketDataStreamer.GetRequestWithWknAndStockId(_accessToken, "710000", "TRG");

            CancellationTokenSource tokenSource1 = new CancellationTokenSource(); // <3>
            Task tickerTask1 = marketDataStreamer.Stream( // <4>
               marketDataRequestTRG, 
                reply => Console.WriteLine("Source (1): " + reply), // <5>
                tokenSource1.Token);

            SecurityMarketDataRequest marketDataRequestOTC = // <6>
                marketDataStreamer.GetRequestWithWknAndStockId(_accessToken, "710000", "OTC");

            CancellationTokenSource tokenSource2 = new CancellationTokenSource();
            Task tickerTask2 = marketDataStreamer.Stream( // <7>
               marketDataRequestOTC,
                reply => Console.WriteLine("Source (2): " + reply), 
                tokenSource2.Token);

            await Task.Delay(10000); // <8>
            tokenSource1.Cancel(); // <9>

            await Task.Delay(20000);
            tokenSource2.Cancel();
        }
        // end::StreamMarketDataAsync[]


        // tag::Login[]
        public string Login(string secret)
        {
            var accessServiceClient =
                new AccessService.AccessServiceClient(channel); // <1>
            var loginRequest = new LoginRequest()
            {
                Secret = secret // <2>
            };
            var loginReply = accessServiceClient.Login(loginRequest); // <3>

            if (loginReply.Error != null)
            {
                // Error handling
                Console.WriteLine(loginReply.Error);
                return null;
            }
            else
            {
                return loginReply.AccessToken; // <4>
            }
        }
        // end::Login[]

        // tag::Logout[]
        public void Logout()
        {
            if (this._accessToken != "")
            {
                var accessServiceClient =
                    new AccessService.AccessServiceClient(channel); // <1>
                var logoutRequest = new LogoutRequest()
                {
                    AccessToken = _accessToken // <2>
                };
                accessServiceClient.Logout(logoutRequest); // <3>
            }
        }
        // end::Logout[]


        // tag::GetTradingAccounts[]
        public TradingAccounts GetTradingAccounts()
        {
            var accountServiceClient =
                new AccountService.AccountServiceClient(channel); // <1>
            var accessToken = new AccessTokenRequest()
            {
                AccessToken = this._accessToken // <2>
            };
            return accountServiceClient.GetTradingAccounts(accessToken); // <3>
        }
        // end::GetTradingAccounts[]

        // tag::StreamTradingAccountInformation[]
        public void StreamTradingAccountInformation()
        {
            TradingAccounts tradingAccounts = GetTradingAccounts(); // <1>
            var accountServiceClient = // <2>
                new AccountService.AccountServiceClient(channel); 
            AccountInformationStreamer streamer = // <3>
                new AccountInformationStreamer(accountServiceClient);
            foreach (var account in tradingAccounts.Accounts) // <4>
            {
                var tradingAccountRequest = new TradingAccountRequest()
                {
                    AccessToken = _accessToken,
                    TradingAccount = account
                };
                Task task = streamer.Stream(tradingAccountRequest, // <5>
                    accountInfomation =>
                    {
                        TradingAccount recievedAccount = accountInfomation.Account; // <6>
                        double balance = accountInfomation.Balance;
                        // ...
                        Console.WriteLine(accountInfomation);
                    });
            }
        }
        // end::StreamTradingAccountInformation[]


        // tag::StreamTradingAccountInformationWithStop[]
        public void StreamTradingAccountInformationWithStop(TradingAccount account)
        {
            var accountServiceClient = // <1>
                new AccountService.AccountServiceClient(channel);
            AccountInformationStreamer streamer = // <2>
                new AccountInformationStreamer(accountServiceClient);
            var tradingAccountRequest = new TradingAccountRequest()
            {
                    AccessToken = _accessToken,
                    TradingAccount = account
            };
            CancellationTokenSource tokenSource = new CancellationTokenSource(); // <3>

            Task task = streamer.Stream(tradingAccountRequest, // <4>
                accountInfomation =>
                {
                        TradingAccount recievedAccount = accountInfomation.Account; // <5>
                        double balance = accountInfomation.Balance;
                        // ...
                        Console.WriteLine(accountInfomation);
                }, 
                tokenSource.Token); // <6>

            tokenSource.Cancel(); // <7>
        }
        // end::StreamTradingAccountInformationWithStop[]

        // tag::GetTradingAccountsAsync[]
        public Task<TradingAccounts> GetTradingAccountsAsync()
        {
            var accountServiceClient =
                new AccountService.AccountServiceClient(channel); // <1>
            var accountStreamer = new AccountsStreamer(accountServiceClient); // <2>
            CancellationTokenSource tokenSorce = new CancellationTokenSource(); // <3>
            var accessTokenRequest = new AccessTokenRequest()
            {
                AccessToken = _accessToken
            };
            var backgroundTask = accountStreamer.Process(accessTokenRequest, 
                accounts => Console.WriteLine(accounts), // <4>
                tokenSorce.Token); 

            return backgroundTask; // <5>
        }
        // end::GetTradingAccountsAsync[]


        // tag::StreamDepotData[]
        private async Task StreamDepotData()
        {
            var tradingAccounts = GetTradingAccounts(); // <1>
            DepotStreamer depotStreamer = new DepotStreamer(depotServiceClient); // <2>
            List<CancellationTokenSource> cancellationTokens = new List<CancellationTokenSource>();
            foreach (var tradingAccount in tradingAccounts.Accounts) // <3>
            {
                var tradingAccountRequest = new TradingAccountRequest() // <4>
                {
                    AccessToken = _accessToken,
                    TradingAccount = tradingAccount
                };

                CancellationTokenSource tokenSource = new CancellationTokenSource(); // <5>
                cancellationTokens.Add(tokenSource);
                var stream = depotStreamer.Stream(tradingAccountRequest, // <6>
                    depotEntries => Console.WriteLine( // <7>
                        "Account:" + depotEntries.Account + ", Data:" + depotEntries),
                    tokenSource.Token);
            }

            await Task.Delay(10000); // <8>

            foreach (var cancellationTokenSource in cancellationTokens) // <9>
            {
                cancellationTokenSource.Cancel();
            }
        }
        // end::StreamDepotData[]

        // tag::GetSecurityCode[]
        public SecurityCode GetSecurityCode(string code, SecurityCodeType codeType)
        {
            return new SecurityCode() // <1>
            {
                Code = code, // <2>
                CodeType = codeType // <3>
            };
        }
        // end::GetSecurityCode[]

        // tag::GetStockExchanges[]
        public StockExchangeDescriptions GetStockExchanges()
        {
            AccessTokenRequest request = new AccessTokenRequest() // <1>
            {
                AccessToken = _accessToken
            };
            return stockExchangeServiceClient.GetStockExchanges(request); // <2>
        }
        // end::GetStockExchanges[]


        // tag::GetBaadaStockExchangeInformation[]
        public StockExchangeDescription GetBaadaStockExchangeInformation()
        {
            StockExchange stockExchange = new StockExchange() // <1>
            {
                Id = "OTC",
                Issuer = "BAAD"
            };
            StockExchangeRequest request = new StockExchangeRequest() // <2>
            {
                AccessToken = _accessToken,
                StockExchange = stockExchange
            };
            return stockExchangeServiceClient.GetStockExchange(request); // <3>
        }
        // end::GetBaadaStockExchangeInformation[]

        // tag::RequestSecurityInfo[]
        public SecurityInfoReply RequestSecurityInfo(string code, SecurityCodeType codeType)
        {
            SecurityCode securityCode = new SecurityCode() // <1>
            {
                Code = code,
                CodeType = codeType
            };

            SecurityInfoRequest request = new SecurityInfoRequest() // <2>
            {
                AccessToken = _accessToken,
                SecurityCode = securityCode
            };

            return securityServiceClient.GetSecurityInfo(request); // <3>
        }
        // end::RequestSecurityInfo[]

        // tag::GetSecurityWithStockExchange[]
        public SecurityWithStockExchange GetSecurityWithStockExchange(string code, SecurityCodeType codeType,
            string stockExchange, string issuer = "")
        {
            return new SecurityWithStockExchange() // <1>
            {
                SecurityCode = new SecurityCode() // <2>
                {
                    Code = code,
                    CodeType = codeType
                },
                StockExchange = new StockExchange() // <3>
                {
                    Id = stockExchange,
                    Issuer = issuer == null ? "" : issuer
                }
            };
        }
        // end::GetSecurityWithStockExchange[]

        // tag::StreamMarketData[]
        public CancellationTokenSource StreamMarketData(string code, SecurityCodeType codeType,
            string stockExchange, string currency)
        {
            SecurityWithStockExchange securityWithStockExchange = new SecurityWithStockExchange() // <1>
            {
                SecurityCode = new SecurityCode() // <2>
                {
                    Code = code,
                    CodeType = codeType
                },
                StockExchange = new StockExchange() // <3>
                {
                    Id = stockExchange
                }
            };

            SecurityMarketDataRequest request = new SecurityMarketDataRequest() // <4>
            {
                AccessToken = _accessToken,
                SecurityWithStockexchange = securityWithStockExchange,
                Currency = currency == null ? "" : currency
            };

            MarketDataStreamer marketDataStreamer =
                new MarketDataStreamer(securityServiceClient); // <5>

            CancellationTokenSource tokenSource = new CancellationTokenSource(); // <6>
            var task = marketDataStreamer.Stream( // <7>
                request,
                reply => Console.WriteLine("Process tick: " + reply), // <8>
                tokenSource.Token);

            return tokenSource; // <9>
        }
        // end::StreamMarketData[]

        // tag::GetSecurityPriceHistory[]
        public SecurityPriceHistoryReply GetSecurityPriceHistory(string code, SecurityCodeType codeType,
            string stockExchange, int days, TimeResolution timeResolution)
        {
            var securityWithStockExchange = // <1>
                GetSecurityWithStockExchange(code, codeType, stockExchange);
            SecurityPriceHistoryRequest request = new SecurityPriceHistoryRequest() // <2>
            {
                AccessToken = _accessToken,
                SecurityWithStockexchange = securityWithStockExchange,
                Days = days,
                TimeResolution = timeResolution
            };

            return securityServiceClient.GetSecurityPriceHistory(request); // <3>
        }
        // end::GetSecurityPriceHistory[]

        // tag::StreamCurrencyRate[]
        public CancellationTokenSource StreamCurrencyRate(string currencyFrom, string currencyTo)
        {
            CurrencyRateRequest request = new CurrencyRateRequest() // <1>
            {
                AccessToken = _accessToken,
                CurrencyFrom = currencyFrom,
                CurrencyTo = currencyTo
            };

            CurrencyRateStreamer currencyRateStreamer =
                new CurrencyRateStreamer(securityServiceClient); // <2>

            CancellationTokenSource tokenSource = new CancellationTokenSource(); // <3>
            var task = currencyRateStreamer.Stream( // <4>
                request,
                reply => Console.WriteLine("Process tick: " + reply), // <5>
                tokenSource.Token);

            return tokenSource; // <6>
        }
        // end::StreamCurrencyRate[]

        // tag::GetStockExchange[]
        public StockExchange GetStockExchange(string id, string issuer = "")
        {
            return new StockExchange()
            {
                Id = id,
                Issuer = issuer == null ? "" : issuer
            };
        }
        // end::GetStockExchange[]


        // tag::RequestQuotesDirect[]
        public void RequestQuotesDirect(string code, SecurityCodeType codeType)
        {
            QuoteRequest request = new QuoteRequest() // <1>
            {
                AccessToken = _accessToken,
                OrderType = OrderType.Buy,
                SecurityCode = new SecurityCode()
                {
                    Code = code,
                    CodeType = codeType
                },
                Amount = 10,
                StockExchanges =
                {
                    GetStockExchange("BAA"), // <2>
                    GetStockExchange("OTC"),
                    GetStockExchange("TRG")
                }
            };

            QuoteReply quoteReply = orderServiceClient.GetQuote(request); // <3>

            if (quoteReply.Error == null) // <4>
            {
                foreach (var quoteEntry in quoteReply.PriceEntries) // <5>
                {
                    if (quoteEntry.Error == null) // <6>
                    {
                        Console.Write("Quote: " + quoteEntry);
                    }
                    else
                    {
                        Console.Write("Error: "+quoteEntry.Error);
                    }
                    
                }
            }
        }
        // end::RequestQuotesDirect[]

        // tag::RequestQuotesAsynchronically[]
        public QuoteReply RequestQuotesAsynchronically(string code, SecurityCodeType codeType)
        {
            SecurityCode securityCode = new SecurityCode() // <1>
            {
                Code = code,
                CodeType = codeType
            };
            StockExchange[] stockExchanges = // <2>
            {
                GetStockExchange("BAA"), 
                GetStockExchange("OTC"),
                GetStockExchange("TRG")
            };

            CountdownEvent countdown = new CountdownEvent(3); // <3>
            object resultsLock = new object();
            QuoteReply reply = null;

            foreach (var stockExchange in stockExchanges) // <4>
            {
                QuoteRequest request = new QuoteRequest() // <5>
                {
                    AccessToken = _accessToken,
                    OrderType = OrderType.Buy,
                    Amount = 10,
                    SecurityCode = securityCode,
                    StockExchanges = {stockExchange}
                };

                var taskAwaiter = orderServiceClient.GetQuoteAsync(request).GetAwaiter(); // <6>
                taskAwaiter.OnCompleted(() => // <7>
                {
                    QuoteReply currentReply = taskAwaiter.GetResult(); // <8>
                    lock (resultsLock)
                    {
                        if (currentReply != null && 
                            currentReply.Error == null && 
                            currentReply.PriceEntries.Count>0)
                        {
                            // Find lowest buy price
                            if (reply == null ||
                                currentReply.PriceEntries[0].BuyPrice < reply.PriceEntries[0].BuyPrice)
                            {
                                reply = currentReply;
                            }
                        }
                    }
                    countdown.Signal();
                });
            }
            countdown.WaitHandle.WaitOne(); // <9>
            return reply;
        }
        // end::RequestQuotesAsynchronically[]

        // tag::AcceptQuote[]
        public OrderReply AcceptQuote(string code, SecurityCodeType codeType)
        {
            var stockExchange = GetStockExchange("OTC");
            var securityCode = new SecurityCode()
            {
                Code = code,
                CodeType = codeType
            };
            QuoteRequest request = new QuoteRequest() // <1>
            {
                AccessToken = _accessToken,
                OrderType = OrderType.Buy,
                SecurityCode = securityCode,
                Amount = 10,
                StockExchanges =
                {
                    stockExchange,
                }
            };

            QuoteReply quoteReply = orderServiceClient.GetQuote(request); // <2>
            if (quoteReply != null &&
                quoteReply.Error != null) 
            {
                var quoteReplyPriceEntry = quoteReply.PriceEntries[0];
                if (quoteReplyPriceEntry.Error != null && // <3>
                    quoteReplyPriceEntry.QuoteReference != null)
                {
                    AcceptQuoteRequest acceptQuoteRequest = new AcceptQuoteRequest() // <4>
                    {
                        AccessToken = _accessToken,
                        AccountNumber = "123456789",
                        OrderType = OrderType.Buy,
                        SecurityWithStockexchange = new SecurityWithStockExchange()
                        {
                            SecurityCode = securityCode,
                            StockExchange = stockExchange
                        },
                        Amount = 10,
                        Validation = Validation.ValidateOnly,
                        Limit = quoteReplyPriceEntry.BuyPrice, // <5>
                        QuoteReference = quoteReplyPriceEntry.QuoteReference // <6>
                    };
                    return orderServiceClient.AcceptQuote(acceptQuoteRequest); // <7>
                }
            }
            return null;
        }
        // end::AcceptQuote[]


        // tag::AcceptQuoteWithCostsRequest[]
        public OrderReply AcceptQuoteWithCostsRequest(string code, SecurityCodeType codeType)
        {
            var stockExchange = GetStockExchange("OTC");
            var securityCode = new SecurityCode()
            {
                Code = code,
                CodeType = codeType
            };
            AcceptQuoteRequest acceptQuoteCostsRequest = new AcceptQuoteRequest() // <1>
            {
                AccessToken = _accessToken,
                AccountNumber = "123456789",
                OrderType = OrderType.Buy,
                SecurityWithStockexchange = new SecurityWithStockExchange()
                {
                    SecurityCode = securityCode,
                    StockExchange = stockExchange
                },
                Validation = Validation.TotalCostsOnly, // <2>
                Amount = 10, // <3>
            };
            OrderReply costsReply = orderServiceClient.AcceptQuote(acceptQuoteCostsRequest); // <4>

            if (costsReply.Error != null) // <5>
            {
                Console.WriteLine(costsReply.Error);
                return costsReply;
            }

            // Sometimes later
            QuoteRequest request = new QuoteRequest() // <6>
            {
                AccessToken = _accessToken,
                OrderType = OrderType.Buy,
                SecurityCode = securityCode,
                Amount = 10,
                StockExchanges =
                {
                    stockExchange,
                }
            };


            QuoteReply quoteReply = orderServiceClient.GetQuote(request); // <7>
            if (quoteReply != null &&
                quoteReply.Error != null)
            {
                var quoteReplyPriceEntry = quoteReply.PriceEntries[0];
                if (quoteReplyPriceEntry.Error != null && 
                    quoteReplyPriceEntry.QuoteReference != null)
                {
                    AcceptQuoteRequest acceptQuoteRequest = new AcceptQuoteRequest() // <8>
                    {
                        AccessToken = _accessToken,
                        AccountNumber = "123456789",
                        OrderType = OrderType.Buy,
                        SecurityWithStockexchange = new SecurityWithStockExchange()
                        {
                            SecurityCode = securityCode,
                            StockExchange = stockExchange
                        },
                        Amount = 10,
                        Validation = Validation.WithoutValidation, // <9>
                        Limit = quoteReplyPriceEntry.BuyPrice, // <10>
                        QuoteReference = quoteReplyPriceEntry.QuoteReference // <11>
                    };
                    return orderServiceClient.AcceptQuote(acceptQuoteRequest); // <12>
                }
            }
            return null;
        }
        // end::AcceptQuoteWithCostsRequest[]

        // tag::GetTodayDate[]
        public Date GetTodayDate()
        {
            DateTime localDate = DateTime.Now;
            return new Date()
            {
                Day = localDate.Day,
                Month = localDate.Month,
                Year = localDate.Year
            };
        }
        // end::GetTodayDate[]

        // tag::AddOrder[]
        public OrderReply AddOrder(TradingAccount tradingAccount)
        {
            if (tradingAccount != null)
            {
                AddOrderRequest request = new AddOrderRequest()
                {
                    AccessToken = _accessToken,
                    AccountNumber = tradingAccount.AccountNumber, // <1>
                    SecurityWithStockexchange = new SecurityWithStockExchange() // <2>
                    {
                        SecurityCode = new SecurityCode()
                        {
                            Code = "710000",
                            CodeType = SecurityCodeType.Wkn
                        },
                        StockExchange = GetStockExchange("OTC")
                    },
                    Validation = Validation.ValidateOnly, // <3>
                    Amount = 1,
                    OrderModel = OrderModel.Limit, // <4>
                    Limit = 1,
                    OrderType = OrderType.Buy, // <5>
                    CashQuotation = CashQuotation.Kassa,
                    ValidityDate = GetTodayDate() // <6>
                };
                return orderServiceClient.AddOrder(request); // <7>
            }

            return null;
        }
        // end::AddOrder[]

        // tag::AddOrderWithCostsRequest[]
        public OrderReply AddOrderWithCostsRequest(TradingAccount tradingAccount)
        {
            if (tradingAccount != null)
            {
                AddOrderRequest costsRequest = new AddOrderRequest() // <1>
                {
                    AccessToken = _accessToken,
                    AccountNumber = tradingAccount.AccountNumber, 
                    SecurityWithStockexchange = new SecurityWithStockExchange() // <2>
                    {
                        SecurityCode = new SecurityCode()
                        {
                            Code = "710000",
                            CodeType = SecurityCodeType.Wkn
                        },
                        StockExchange = GetStockExchange("OTC")
                    },
                    Validation = Validation.TotalCostsOnly, // <3>
                    Amount = 1,
                    OrderModel = OrderModel.Limit,
                    Limit = 1,
                    OrderType = OrderType.Buy, // <4>
                    CashQuotation = CashQuotation.Kassa,
                    ValidityDate = GetTodayDate() 
                };

                OrderReply costsReply = orderServiceClient.AddOrder(costsRequest); // <5>
                if (costsReply.Error != null) // <6>
                {
                    Console.WriteLine(costsReply.Error);
                    return costsReply;
                }

                AddOrderRequest request = new AddOrderRequest() // <7>
                {
                    AccessToken = _accessToken,
                    AccountNumber = tradingAccount.AccountNumber, 
                    SecurityWithStockexchange = new SecurityWithStockExchange() 
                    {
                        SecurityCode = new SecurityCode()
                        {
                            Code = "710000",
                            CodeType = SecurityCodeType.Wkn
                        },
                        StockExchange = GetStockExchange("OTC")
                    },
                    Validation = Validation.WithoutValidation, // <8>
                    Amount = 1,
                    OrderModel = OrderModel.Limit, 
                    Limit = 1,
                    OrderType = OrderType.Buy, 
                    CashQuotation = CashQuotation.Kassa,
                    ValidityDate = GetTodayDate() 
                };
                return orderServiceClient.AddOrder(request); // <9>
            }

            return null;
        }
        // end::AddOrderWithCostsRequest[]

        // tag::ChangeOrder[]
        public OrderReply ChangeOrder(Order order, TradingAccount tradingAccount)
        {
            if (order != null && tradingAccount != null)
            {
                string orderNumber = order.OrderNumber;
                if (orderNumber != null && !orderNumber.Equals(""))
                {
                    ChangeOrderRequest request = new ChangeOrderRequest() // <1>
                    {
                        AccessToken = _accessToken,
                        AccountNumber = tradingAccount.AccountNumber, // <2>
                        OrderNumber = orderNumber, // <3>
                        Validation = Validation.ValidateOnly, // <4>
                        OrderModel = OrderModel.Limit, // <5>
                        Limit = 2 
                    };

                    return orderServiceClient.ChangeOrder(request); // <6>
                }
            }
            return null;
        }
        // end::ChangeOrder[]

        // tag::CancelOrder[]
        public OrderReply CancelOrder(Order order, TradingAccount tradingAccount)
        {
            if (order != null && tradingAccount != null)
            {
                string orderNumber = order.OrderNumber;
                if (orderNumber != null && !orderNumber.Equals(""))
                {
                    CancelOrderRequest request = new CancelOrderRequest() // <1>
                    {
                        AccessToken = _accessToken,
                        AccountNumber = tradingAccount.AccountNumber, // <2>
                        OrderNumber = orderNumber, // <3>
                        Validation = Validation.ValidateOnly, // <4>
                    };

                    return orderServiceClient.CancelOrder(request); // <5>
                }
            }
            return null;
        }
        // end::CancelOrder[]

        // tag::ActivateOrder[]
        public OrderReply ActivateOrder(Order order, TradingAccount tradingAccount)
        {
            if (order != null && tradingAccount != null)
            {
                string orderNumber = order.OrderNumber;
                if (orderNumber != null && !orderNumber.Equals(""))
                {
                    ActivateOrderRequest request = new ActivateOrderRequest() // <1>
                    {
                        AccessToken = _accessToken,
                        AccountNumber = tradingAccount.AccountNumber, // <2>
                        OrderNumber = orderNumber, // <3>
                        Validation = Validation.ValidateOnly, // <4>
                    };

                    return orderServiceClient.ActivateOrder(request); // <5>
                }
            }
            return null;
        }
        // end::ActivateOrder[]

        // tag::DeactivateOrder[]
        public OrderReply DeactivateOrder(Order order, TradingAccount tradingAccount)
        {
            if (order != null && tradingAccount != null)
            {
                string orderNumber = order.OrderNumber;
                if (orderNumber != null && !orderNumber.Equals(""))
                {
                    DeactivateOrderRequest request = new DeactivateOrderRequest() // <1>
                    {
                        AccessToken = _accessToken,
                        AccountNumber = tradingAccount.AccountNumber, // <2>
                        OrderNumber = orderNumber, // <3>
                        Validation = Validation.ValidateOnly, // <4>
                    };

                    return orderServiceClient.DeactivateOrder(request); // <5>
                }
            }
            return null;
        }
        // end::DeactivateOrder[]

        // tag::GetOrderCosts[]
        OrderReply GetOrderCosts(TradingAccount tradingAccount)
        {
            AddOrderRequest request = new AddOrderRequest()
            {
                AccessToken = _accessToken,
                AccountNumber = tradingAccount.AccountNumber, // <1>
                SecurityWithStockexchange = new SecurityWithStockExchange() // <2>
                {
                    SecurityCode = new SecurityCode()
                    {
                        Code = "710000",
                        CodeType = SecurityCodeType.Wkn
                    },
                    StockExchange = GetStockExchange("OTC")
                },
                Validation = Validation.ValidateWithTotalCosts, // <3>
                // Validation = Validation.ValidateWithDetailCosts,
                Amount = 1,
                OrderModel = OrderModel.Limit, // <4>
                Limit = 1,
                OrderType = OrderType.Buy, // <5>
                CashQuotation = CashQuotation.Kassa,
                ValidityDate = GetTodayDate() // <6>
            };
            OrderReply orderReply = orderServiceClient.AddOrder(request); // <7>

            if (orderReply.Error != null)
            {
                OrderCosts orderCosts = orderReply.OrderCosts; // <8>
                double estimatedCosts = orderCosts.EstimatedTotalCosts;
                foreach (var categorieCost in orderCosts.CategorieCosts) // <9>
                {
                    double totalSubAbsolute = categorieCost.TotalSumAbsolute;
                    // ...
                }
            }

            Console.WriteLine("Costs reply:"+orderReply);
            return orderReply;
        }
        // end::GetOrderCosts[]

        // tag::StreamOrders[]
        public CancellationTokenSource StreamOrders(TradingAccount tradingAccount)
        {
            TradingAccountRequest request = new TradingAccountRequest() // <1>
            {
                AccessToken = _accessToken,
                TradingAccount = tradingAccount 
            };

            OrdersStreamer ordersStreamer = new OrdersStreamer(orderServiceClient); // <2>

            CancellationTokenSource tokenSource = new CancellationTokenSource(); // <3>

            var task =ordersStreamer.Stream( // <4>
                request,
                reply => Console.WriteLine("Process orders: " + reply), // <5>
                tokenSource.Token);

            return tokenSource; // <6>
        }
        // end::StreamOrders[]

        // tag::StreamOrderBook[]
        public CancellationTokenSource StreamOrderBook(string code, SecurityCodeType codeType,
            string stockExchange, string currency)
        {
            SecurityWithStockExchange securityWithStockExchange = new SecurityWithStockExchange() // <1>
            {
                SecurityCode = new SecurityCode() // <2>
                {
                    Code = code,
                    CodeType = codeType
                },
                StockExchange = new StockExchange() // <3>
                {
                    Id = stockExchange
                }
            };

            SecurityOrderBookRequest request = new SecurityOrderBookRequest() // <4>
            {
                AccessToken = _accessToken,
                SecurityWithStockexchange = securityWithStockExchange,
                Currency = currency == null ? "" : currency
            };

            OrderBookStreamer orderBookStreamer =
                new OrderBookStreamer(securityServiceClient); // <5>

            CancellationTokenSource tokenSource = new CancellationTokenSource(); // <6>
            var task = orderBookStreamer.Stream( // <7>
                request,
                reply => Console.WriteLine("Process tick: " + reply), // <8>
                tokenSource.Token);

            return tokenSource; // <9>
        }
        // end::StreamOrderBook[]

        // tag::ProcessMarketDataExample[]
        public void ProcessMarketDataExample()
        {
            CancellationTokenSource tokenSource = // <1>
                StreamMarketData("710000", SecurityCodeType.Wkn, "OTC", "");

            System.Threading.Thread.Sleep(1000); // <2>

            tokenSource.Cancel(); // <3>
        }
        // end::ProcessMarketDataExample[]

        // tag::ProcessOrderBookExample[]
        public void ProcessOrderBookExample()
        {
            CancellationTokenSource tokenSource = // <1>
                StreamOrderBook("710000", SecurityCodeType.Wkn, "ETR", "");

            System.Threading.Thread.Sleep(1000); // <2>

            tokenSource.Cancel(); // <3>
        }
        // end::ProcessOrderBookExample[]

        // tag::Main[]
        public Program(string host, int port, string trustCertificate, string secret)
        {
            SslCredentials channelCredentials = new SslCredentials(trustCertificate);  // <1>
            channel = new Channel(host + ":" + port, channelCredentials); // <2> 

            this._accessToken = Login(secret);
            if (this._accessToken == null)
            {
                throw new RuntimeBinderException("Error by login");
            }

            stockExchangeServiceClient = new StockExchangeService.StockExchangeServiceClient(channel);
            depotServiceClient = new DepotService.DepotServiceClient(channel);
            securityServiceClient = new SecurityService.SecurityServiceClient(channel);
            orderServiceClient = new OrderService.OrderServiceClient(channel);
        }

        public void Dispose()
        {
            Logout();
            channel?.ShutdownAsync().Wait(); // <3>
        }

        public static void Main(string[] args)
        {
            if (args.Length<2)
            {
                Console.WriteLine("Please generate key in AT and" +
                    "put location of the trust certificate and secret as an argument");
                return;
            }

            string trustCertificate = File.ReadAllText(args[0]); // <4>
            using (Program program = new Program("localhost", 40443, 
                trustCertificate, args[1])) { // <5>
                // end::Main[]
                var tradingAccounts = program.GetTradingAccounts();

                /*
                var orderReply = program.GetOrderCosts(tradingAccounts.Accounts[0]);

                if (orderReply != null)
                {
                    orderReply = program.AddOrder(tradingAccounts.Accounts[0]);
                }

                if (orderReply != null && orderReply.Order != null && orderReply.Account != null)
                {
                    orderReply = program.ChangeOrder(orderReply.Order, orderReply.Account);
                }

                if (orderReply != null && orderReply.Order != null && orderReply.Account != null)
                {
                    // Pro only
                    orderReply = program.DeactivateOrder(orderReply.Order, orderReply.Account);
                }

                if (orderReply != null && orderReply.Order != null && orderReply.Account != null)
                {
                    // Pro only
                    orderReply = program.ActivateOrder(orderReply.Order, orderReply.Account);
                }

                if (orderReply != null && orderReply.Order != null && orderReply.Account != null)
                {
                    orderReply = program.CancelOrder(orderReply.Order, orderReply.Account);
                }
                */

                QuoteReply reply = program.RequestQuotesAsynchronically("710000", SecurityCodeType.Wkn);
                Console.WriteLine("->" + reply);


                Console.WriteLine("->"+program.GetSecurityPriceHistory("710000", SecurityCodeType.Wkn, "OTC", 10, TimeResolution.Day));

                program.RequestQuotesDirect("710000", SecurityCodeType.Wkn);


                var cancellationTokenSource = program.StreamOrders(tradingAccounts.Accounts[0]);

                program.StreamCurrencyRate("EUR", "USD");
                program.ProcessOrderBookExample();
                program.ProcessMarketDataExample();

                program.StreamDepotData().Wait();

                var stockExchangeDescriptions = program.GetStockExchanges();
                var baadaStockExchangeInformation = program.GetBaadaStockExchangeInformation();

                program.StreamMarketDataAsync().Wait();
                // tag::Main[]

                // ...

            }
        }

    }
    // end::Main[]
}
